from typing import Generic, TypeVar
from unittest import mock

from django.utils import timezone

from olympus.base import OlympusCollector

C = TypeVar("C", bound=OlympusCollector)


class HelperTestCollectorMixin(Generic[C]):
    """Helper to test sub-classes of OlympusCollector (e.g occ.py)"""

    COLLECTOR_CLS: type[C]

    es_mock: mock.MagicMock
    collector: C
    test_now = timezone.datetime(year=2020, month=3, day=23, tzinfo=timezone.utc)
    test_now_patch: mock._patch
    test_now_mock: mock.MagicMock

    def setUp(self) -> None:
        self.test_now_patch = mock.patch("django.utils.timezone.now", return_value=self.test_now)
        self.test_now_mock = self.test_now_patch.start()
        self.es_mock = mock.MagicMock()
        self.collector = self.COLLECTOR_CLS(es=self.es_mock)

    def tearDown(self) -> None:
        self.test_now_patch.stop()



from django.apps import AppConfig
from django.conf import settings

APP_SETTINGS = {
    'ELASTICSEARCH_URL': '127.0.0.1',
    'ELASTICSEARCH_VERIFY_CERTS': True,
}


class OlympusConfig(AppConfig):
    name = 'olympus'

    def ready(self):
        super().ready()
        for k, v in APP_SETTINGS.items():
            _k = f'{self.name.upper()}_{k}'
            if not hasattr(settings, _k):
                setattr(settings, _k, v)
        self.module.autodiscover()



import re
import logging
import tqdm
from logbasecommand.base import LogBaseCommand, CommandError

from olympus.base import collectors


class Command(LogBaseCommand):
    help = 'Push data to Olympus (ElasticSearch).'

    def add_arguments(self, parser):
        parser.add_argument('app_class', type=str, nargs='*', help='App or specific class to process')
        parser.add_argument('--no-progress', action='store_true', help='Disable progress-bar')
        parser.add_argument('--test', action='store_true', help='Run collector without pushing results to ES, use verbosity to view data')

    def print_collectors(self):
        for _, _, collector in collectors():
            self.stdout.write(f'{collector.name()} ({collector().get_index_name()})')

    def find_chosen(self, cols):
        matches = []
        for c in cols:
            if 0 <= c.find('*') < len(c) - 1:
                raise CommandError('wildcard only supported at the end', c)
            if c[-1] == '*':
                matches.append(re.compile(re.escape(c[:-1]) + r'.*'))
            elif '.' not in c:
                matches.append(re.compile(re.escape(c) + r'\..*'))
            else:
                matches.append(re.compile(re.escape(c) + r'$'))

        chosen = []
        for _, cname, cclass in collectors():
            for _m in matches:
                if _m.match(cname):
                    chosen.append((cname, cclass))

        if not chosen:
            raise CommandError('no valid collectors specified')
        return chosen
    
    def _set_es_logger_level(self):
        if self.verbosity == 2:
            logging.getLogger('elasticsearch').setLevel(logging.INFO)
        elif self.verbosity > 2:
            logging.getLogger('elasticsearch').setLevel(logging.DEBUG)

    def handle(self, *args, **options):
        cols = options['app_class']

        # do not allow running all collectors (unless explicitly enumerated)
        # some might not be supposed to run "wild"
        if not cols:
            return self.print_collectors()

        chosen = self.find_chosen(cols)

        self._set_es_logger_level()

        if options['no_progress']:
            self.log(f'Matched {len(chosen)} collectors')

        errors = []
        ind = 0

        for cname, cclass in chosen:
            self.log_debug(f'Running {cclass.name()}...')
            col = cclass()
            col.logger.setLevel(self.logger.getEffectiveLevel())
            estimated_count = None if options['no_progress'] else col.estimated_count()
            push = col.fake_push if options['test'] else col.push
            ind += 1

            with tqdm.tqdm(
                desc=f'{cclass.name()} {ind}/{len(chosen)}', disable=options['no_progress'], total=estimated_count
            ) as pbar:
                error_bar = ErrorBar(pbar)
                try:
                    stats = push(stats_cb=error_bar.update)
                    if options['no_progress']:
                        self.log(f'{cname} pushed {stats[0]} records')
                    if stats[1]:
                        errors.append(cname)
                        self.log_error('%s failed: %s', cname, stats[1])
                except Exception:
                    errors.append(cname)
                    self.log_exception('%s failed', cname)

        if errors:
            raise CommandError('These collectors failed', errors)


class ErrorBar:
    def __init__(self, pbar):
        self.err = 0
        self.pbar = pbar

    def update(self, ok, _):
        self.pbar.update(1)  # NOSONAR python:S1515
        if not ok:
            self.err += 1  # NOSONAR python:S1515
            self.pbar.set_postfix(err=self.err)



import logging
from typing import Any, Callable, Iterator, Optional

from django.conf import settings
from django.utils import timezone
from elasticsearch import Elasticsearch, helpers


_collectors: list[tuple[str, str, type["OlympusCollector"]]] = []


class ElasticSearchClient(Elasticsearch):
    def __init__(self, hosts=None, **kwargs):
        if hosts is None:
            hosts = [settings.OLYMPUS_ELASTICSEARCH_URL]
        kwargs['verify_certs'] = kwargs.get('verify_certs', settings.OLYMPUS_ELASTICSEARCH_VERIFY_CERTS)

        if "timeout" not in kwargs:
            kwargs["timeout"] = 30

        super().__init__(hosts=hosts, **kwargs)


class OlympusCollector:
    index_name: Optional[str] = None
    index_date_pattern: Optional[str] = None
    chunk_size = 500
    alias_suffix = '-latest'

    @classmethod
    def __init_subclass__(cls):
        # register collectors - use metaclass if py < 3.6 support is required
        _collectors.append((cls.__module__.split('.')[0], cls.name(), cls))
        return super().__init_subclass__()

    def __init__(self, es: Optional[ElasticSearchClient] = None, timestamp: Optional[timezone.datetime] = None):
        self.logger = logging.getLogger(f'occ:{self.__module__}.{self.__class__.__name__}')
        self.es = es if es else ElasticSearchClient()
        if timestamp is None:
            self.timestamp = timezone.now()
        else:
            self.timestamp = timestamp

    def get_index_name(self) -> str:
        # set property for simple index name
        # or just override this method for more complex logic
        index_name = self.__get_raw_index_name()
        if self.index_date_pattern is None:
            return index_name

        return f'{index_name}-{self.timestamp.strftime(self.index_date_pattern)}'

    def __get_raw_index_name(self) -> str:
        return self.index_name or self.name().lower()

    @classmethod
    def name(cls) -> str:
        return f"{cls.__module__.split('.')[0]}.{cls.__name__}"

    def create_index(self):
        self.es.indices.create(index=self.get_index_name(), ignore=400)
        if self.index_date_pattern and self.alias_suffix:
            # Create unique alias for latest index
            alias = self.__get_raw_index_name() + self.alias_suffix
            self.es.indices.delete_alias(index='_all', name=alias, ignore=404)
            self.es.indices.put_alias(index=self.get_index_name(), name=alias)

    def fake_push(self,
        stats_cb: Optional[Callable[[bool, dict[str, Any]], None]] = None,
    ) -> tuple[int, list[dict[str, Any]]]:
        """
        this method collects data but does not actually push it to ES, log.debug it instead

        :param stats_cb: function called with (ok, item) as returned by streaming_bulk on every iteration
        :return: tuple of number of (success, fail) records
        """
        success = 0
        for item in self.__collect():
            success += 1
            if stats_cb is not None:
                stats_cb(True, item)
            self.logger.debug('would push: %s', item)
        return success, []

    def push(
        self,
        stats_cb: Optional[Callable[[bool, dict[str, Any]], None]] = None,
    ) -> tuple[int, list[dict[str, Any]]]:
        """
        :param stats_cb: function called with (ok, item) as returned by streaming_bulk on every iteration
        :return: tuple of number of (success, fail) records
        """
        self.create_index()
        # make streaming_bulk yield successful results so we can count them
        success = 0
        fails = []
        for ok, item in helpers.streaming_bulk(
            self.es, self.__collect(), chunk_size=self.chunk_size, yield_ok=True, raise_on_error=False
        ):
            if not ok and 'delete' in item and item['delete'].get('status') == 404:
                # ignore "not_found" on delete...
                # consider it "ok" for stats purposes
                ok = True
            if stats_cb is not None:
                stats_cb(ok, item)
            if not ok:
                fails.append(item)
            else:
                success += 1
        return success, fails

    def __collect(self) -> Iterator[dict[str, Any]]:
        # add default fields to collected items
        for _i in self.collect():
            if '_type' not in _i:
                _i['_type'] = 'status'
            if '_index' not in _i:
                _i['_index'] = self.get_index_name()
            yield _i

    def collect(self) -> Iterator[dict[str, Any]]:
        raise NotImplementedError()

    def estimated_count(self) -> Optional[int]:
        """
        this should return as fast as possible an estimate of the amount of objects that will be generated by collect()
        this is only used for stats, so if there's no quick count for the collector, return None

        ENFORCED:
        this method should not iterate everything like collect() only to provide a count
        that would severely impact that collector performance (iterating twice...)
        if that is the only way, do not use it, just return None.

        :return: estimated number of objects to be generated or None if impossible to predict
        """
        return None


def collectors() -> list[tuple[str, str, type["OlympusCollector"]]]:
    return _collectors
